package parsedata

import (
	"os"
	"log"
	"sort"
	"encoding/json"
	"ORsearch/consts"
)



type LexSlice [][]uint16

func (a LexSlice) Len() int           { return len(a) }
func (a LexSlice) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
func (a LexSlice) Less(i, j int) bool { return sliceLess(a[i],a[j]) }

// Lexicographical order
func sliceLess(a,b []uint16) bool {
	for n := 0; n < len(a) && n < len(b); n++ {
		if a[n] != b[n] {
			return a[n] < b[n]
		}
	}
	// If they start with the same elements, the shorter is smaller
	return len(a) < len(b)
}



// Build a list of all the already search branches from json format file "inputf"
func ReadOldBranches(inputf string) [][]uint16 {
	var skipbranches = make([][]uint16, 0)
	if inputf == "" {
		return skipbranches
	}

	infile, err := os.Open(inputf) // For reading
	if err != nil {
		log.Fatal(err)
	}
	defer infile.Close()

	jsonin := json.NewDecoder(infile)	

	var curbranch []uint16
	for {
		err := jsonin.Decode(&curbranch)
		if err != nil {
			sort.Sort(LexSlice(skipbranches))	
			return skipbranches
		}
		// We need to append separate copies
		var curbranch2 = make([]uint16, consts.DISTRIBUTIONROWS)
		copy(curbranch2,curbranch)
		skipbranches = append(skipbranches, curbranch2)
	}

	sort.Sort(LexSlice(skipbranches))
	return skipbranches
}



// Check if "branch" is anywhere in "branchlist"
// Branchlist has to be sorted!
func IsInBranchlist(branch []uint16, branchlist [][]uint16) bool {
	if len(branchlist) == 0 {
		return false
	}

	// Logarithmic time search for an element in sorted list
	// 0 <= imin <= icur <= imax < len(branchlist)
	var imin = 0
	var imax = len(branchlist) - 1
	var icur = 0

	for imax > imin + 1 {
		icur = (imax + imin) / 2 
		if sliceLess(branch, branchlist[icur]) {
			imax = icur
		} else {
			imin = icur
		}

	}

	// imin is the only candidate
	// There shouldn't be any length difference, but who knows
	if len(branch) != len(branchlist[imin]) {
		return false
	}

	for i, elem := range branch {
		if branchlist[imin][i] != elem {
			return false
		}
	}
	// If everything is ok, then we've found the element
	return true
}
