package workers

import (
	"fmt"
	"ORsearch/consts"
	"ORsearch/structures"
)



// Search for the matrix with the most rows possible (or with targetlength rows if such a matrix is found during the search)
// The "rows" first rows are fixed to m
//
// Parameters:
// - rows: the current/starting number of rows, which corresponds to the index l = rows-1 in m
// - *m: the current/starting matrix with l rows
// - compatiblepairs: the P_m set (see definition in the paper) 
//	 such that q = compatiblepairs[l][r][i] is the i^th possible q such that (r,q) is in P_m
// - cardinalityQr: cardinalityQr[l][r] is the cardinality of Q_m(r) (see definition in the paper) 
//	 so we know how many q's must be checked in compatiblepairs[l][r]
// - *bestlength: the maximum number of rows of a matrix found during the search
// - *bestmatrix: the matrix with bestlength rows found during the search
// - targetlength: if l <= targetlength <= bestlength, when a matrix with targetlength rows is found
// 	 we do not branch, backtrack one step in the search tree and continue from there
// - *brancheslist: if l <= targetlength <= bestlength, every matrix with targetlength rows is saved here
// - sym: sym[l] defines the symmetry structure of m
// 	 (for a finer definition of sym, see the description of structures.UpdateSymmetry())
func branchSearch(
	rows int, 												// the current number of rows
	m *structures.Matrix, 									// the starting matrix, with l rows
	compatiblepairs [][][]uint16, cardinalityQr [][]int, 	// compatible pairs information
	bestlength *int, bestmatrix *structures.Matrix,			// best result informations
	targetlength int, brancheslist *[][]uint16,				// if some target number of rows is aimed
	sym []uint16) {							// symmetry information

	var l = rows-1 // the index of the rows's row of m

	// Update best matrix and display stuff
	if l > *bestlength {
		// New bestmatrix and bestlength
		bestmatrix.Copy(m) 
		*bestlength = l
		if l >= consts.BESTMATRIXINSTANTREPORT && consts.DISPLAY >= 1 {
			fmt.Println("Intermediate record:", l)
			m.Display()
		}	
	} 

	// If targetlength is reached, we save the current matrix, backtrack one step in the tree and continue the search
	if rows == targetlength {
		var matrixcopy = make([]uint16, rows)
		copy(matrixcopy, m.Elems[:(rows)])
		*brancheslist = append(*brancheslist, matrixcopy)
		return
	}
	// In case we wanted to interrupt the search and return the found matrix, this is what we would do
	// if *bestlength == targetlength {
	// 	if l == targetlength && consts.DISPLAY >= 2 {
	// 		fmt.Println("Target length reached :", l)
	// 		m.Display()
	// 	}
	// 	return
	// }

	// Update the compatible pairs and get a certificate that the branch is susceptible to achieve the CUTATUPDATE level
	var goodbranch = structures.UpdateCompatiblePairs(compatiblepairs, cardinalityQr, m, l)

	if !goodbranch {
		return
	}

	// Explore all branches
	var q uint16
	for i := 0; i < cardinalityQr[l][m.Elems[l]]; i++ {
		q = compatiblepairs[l][m.Elems[l]][i]
		// If symmetry has to be checked (sym[l] != 11...1), check whether q satisfies the symmetry rules
		if structures.IsSym(sym[l]) {
			if structures.IsSymOK(sym[l], q) {
				sym[l+1] = structures.UpdateSymmetry(sym[l], q)
			} else {
				continue
			}
		} else {
			sym[l+1] = sym[l]
		}
		// Branch recursively
		m.Elems[l+1] = q
		branchSearch(rows+1, m, compatiblepairs, cardinalityQr, bestlength, bestmatrix, targetlength, brancheslist, sym)
	}

}



